#!/bin/bash

# (c) 2011, Tjark Weber <tjark.weber@gmx.de>.  All rights reserved.
# Last modified: 2011-01-15

# Requires squolem2 in your $PATH.  This script was tested with
# squolem2 2.02 (compiled Nov 10 2010); later versions might work as
# well.  Squolem2 is currently available from
# http://www.cprover.org/qbv/.

# Requires (relatively small amounts of) free disk space for the
# generated certificates.

# Certificate generation will take a long time.  However, you can
# interrupt this shell script during certificate generation (by
# creating a file called 'stop'), and resume later (by executing the
# script again).  See the source code below for details.

###############################################################################
# Preliminaries
###############################################################################

set -e  # abort this shell script if anything goes wrong
set -u  # forbid use of uninitialized variables

###############################################################################
# Determine remaining benchmarks (in case this script was run and
# interrupted earlier)
###############################################################################

# The log file consists of a sequence of entries, each of which
# adheres to the following format:
#
# ./path/to/benchmark.qdimacs
# [optional output]
# real    x.xxx
# user    x.xxx
# sys     x.xxx
# exit-status xxx
# {true|false|error (reason)|timeout}

if [ -e squolem2.log ]; then
  # only retain complete log entries
  awk 'BEGIN { count=0 } { file[count]=$0; count++ } /^(true|false|error \(.*\)|timeout)$/ { for (i=0; i<count; i++) print file[i]; count=0 }' squolem2.log > squolem2.log.tmp
  mv squolem2.log.tmp squolem2.log
  # number of completed benchmarks
  count_done=`grep -c -e '^\./.*\.qdimacs$' squolem2.log`
  # name of the last completed benchmark
  last_done=`grep -e '^\./.*\.qdimacs$' squolem2.log|tail -n 1`
else
  # let's save some system information
  uname -a > uname-a
  cat /proc/cpuinfo > cpuinfo
  cat /proc/meminfo > meminfo
  squolem2 > squolem2-version

  count_done=0
  last_done=
fi

echo "$count_done benchmark(s) in log, $(( `find . -name '*.qdimacs'|wc -l` - $count_done )) benchmark(s) remaining."

if [ -n "$last_done" ]; then
  echo "Resuming after $last_done"
fi

###############################################################################
# Generating the certificates
###############################################################################

echo "Generating certificates ('touch \"`pwd`/stop\"' to interrupt - DO NOT press Ctrl-C) ..."

# The following command applies squolem2 to each benchmark file, using
# a timeout of $timeout s (for each benchmark), and restricting memory
# usage to $maxmemory KB.  (Feel free to adjust these values.)  This
# will take a long time.  On a multi-core computer, you may wish to
# issue several similar commands in parallel (e.g., in different
# subdirectories of $basename/).

# possible exit status for each squolem2 sub-shell:
# 10   == success, true formula
# 20   == success, false formula
# 137  == killed (ulimit timeout)
# else == some error

TIMEFORMAT=$'real\t%3R\nuser\t%3U\nsys\t%3S'

timeout=600        # 600 s
maxmemory=4194304  # 4 GB

# requires squolem2 in your $PATH
for a in `find . -name '*.qdimacs'|sort`; do
  # exit if the user wants to stop -- Ctrl-C will typically be caught
  # by the squolem2 process, so I'm using this (rather ugly)
  # file-based solution instead
  if [ -e stop ]; then
    rm stop
    echo "Stop file found, exiting. Run this script again to resume."
    exit 1
  fi
  # skip over benchmarks that were completed before
  if [ $count_done -gt 0 ]; then
    count_done=$(( $count_done - 1 ))
    # after skipping $count_done benchmarks, we should be at $last_done
    if [ $count_done -eq 0 ]; then
      if [ "$a" != "$last_done" ]; then
        echo "Error while skipping completed benchmarks (at $a)."
        exit 1
      fi
    fi
    continue
  fi
  echo "$a" >> squolem2.log
  exitstatus=0
  (ulimit -t "$timeout" -v "$maxmemory" && time squolem2 -c "$a") &>> squolem2.log || exitstatus=$?
  echo "exit-status $exitstatus" >> squolem2.log
  case $exitstatus in
    10)  # success (true)
      if [ -e "$a.qbc" ]; then
        echo "true" >> squolem2.log
      else
        echo "error (true formula, but no certificate)" >> squolem2.log
      fi;;
    20)  # success (false)
      if [ -e "$a.qbc" ]; then
        echo "false" >> squolem2.log
      else
        echo "error (false formula, but no certificate)" >> squolem2.log
      fi;;
    137)  # timeout: remove output
      rm -f "$a.qbc.tmp" "$a.qbc"
      echo "timeout" >> squolem2.log;;
    *)  # error: remove output
      rm -f "$a.qbc.tmp" "$a.qbc"
      echo "error (exit status)" >> squolem2.log;;
  esac
done

###############################################################################
# Generating some statistics
###############################################################################

# for each benchmark, compute the sum of user time and system time and
# add it to the squolem2 log file (for further processing of the log),
# unless that has been done before
grep -q -e '^u+s' squolem2.log ||
  {
    awk '{ print } /^user/ { user=$2 } /^sys/ { print "u+s\t" (user + $2) }' squolem2.log > squolem2.log.tmp
    mv squolem2.log.tmp squolem2.log
  }

if [ -e squolem2-statistics.txt ]; then
  echo "File squolem2-statistics.txt exists, skipping statistics generation."
else
  echo "Generating statistics ..."

  # save the number of benchmarks, errors, timeouts and proofs
  # save the disk size of benchmarks and proofs
  # save the solving time
  # Note that size averages are truncated, while time averages are rounded.
  echo 'Benchmarks Errors Timeouts Proofs BenchmarkSize(B) BenchmarkAvgSize(B) ProofSize(B) ProofAvgSize(B) TotalTime(s) ProofTime(s) ProofAvgTime(s)' > squolem2-statistics.txt
  for a in `ls -d */` ""; do
    a="./$a"
    echo "$a" >> squolem2-statistics.txt
    benchmarks=`grep -c -e "^$a.*\.qdimacs$" squolem2.log`
    disk_benchmarks=`find "$a" -name '*.qdimacs'|wc -l`
    if [ "$benchmarks" -ne "$disk_benchmarks" ]; then
      echo "$a: $benchmarks benchmarks in log, $disk_benchmarks benchmarks on disk."
      exit 1
    fi
    errors=`awk 'BEGIN { count=0 } /^\.\/.*\.qdimacs$/ { if (index($0, "'"$a"'") == 1) docount=1; else docount=0 } /^error \(.*\)$/ { count += docount } END { print count }' squolem2.log`
    proofs=`awk 'BEGIN { count=0 } /^\.\/.*\.qdimacs$/ { if (index($0, "'"$a"'") == 1) docount=1; else docount=0 } /^(true|false)$/ { count += docount } END { print count }' squolem2.log`
    disk_proofs=`find "$a" -name '*.qbc'|wc -l`
    if [ "$proofs" -ne "$disk_proofs" ]; then
      echo "$a: $proofs proofs in log, $disk_proofs proofs on disk."
      exit 1
    fi
    timeouts=`awk 'BEGIN { count=0 } /^\.\/.*\.qdimacs$/ { if (index($0, "'"$a"'") == 1) docount=1; else docount=0 } /^timeout$/ { count += docount } END { print count }' squolem2.log`
    disk_timeouts=$(( $benchmarks-$errors-$proofs ))
    if [ "$timeouts" -ne "$disk_timeouts" ]; then
      echo "$a: $timeouts timeouts in log, $disk_timeouts timeouts on disk."
      exit 1
    fi
    benchmark_size=`find "$a" -name '*.qdimacs' -print0|du -bc --files0-from=-|tail -n 1|awk '{ print $1 }'`
    if [ "$benchmarks" -eq 0 ]; then
      benchmark_avgsize="-"
    else
      benchmark_avgsize=$(( $benchmark_size/$benchmarks ))
    fi
    proof_size=`find "$a" -name '*.qbc' -print0|du -bc --files0-from=-|tail -n 1|awk '{ print $1 }'`
    if [ "$proofs" -eq 0 ]; then
      proof_avgsize="-"
    else
      proof_avgsize=$(( $proof_size/$proofs ))
    fi
    total_time=`awk 'BEGIN { time=0 } /^\.\/.*\.qdimacs$/ { if (index($0, "'"$a"'") == 1) docount=1; else docount=0 } /^u\+s/ { if (docount) time += $2 } END { printf("%#.3f", time) }' squolem2.log`
    proof_time=`awk 'BEGIN { time=0 } /^\.\/.*\.qdimacs$/ { if (index($0, "'"$a"'") == 1) docount=1; else docount=0 } /^u\+s/ { thistime=$2 } /^(true|false)$/ { if (docount) time += thistime } END { printf("%#.3f", time) }' squolem2.log`
    if [ "$proofs" -eq 0 ]; then
      proof_avgtime="-"
    else
      proof_avgtime=`awk 'BEGIN { printf ("%#.3f", '"$proof_time"' / '"$proofs"') }'`
    fi
    benchmarks=`printf "%10s" "$benchmarks"`
    errors=`printf "%6s" "$errors"`
    proofs=`printf "%6s" "$proofs"`
    timeouts=`printf "%8s" "$timeouts"`
    benchmark_size=`printf "%16s" "$benchmark_size"`
    benchmark_avgsize=`printf "%19s" "$benchmark_avgsize"`
    proof_size=`printf "%12s" "$proof_size"`
    proof_avgsize=`printf "%15s" "$proof_avgsize"`
    total_time=`printf "%12s" "$total_time"`
    proof_time=`printf "%12s" $proof_time`
    proof_avgtime=`printf "%15s" $proof_avgtime`
    echo "$benchmarks $errors $timeouts $proofs $benchmark_size $benchmark_avgsize $proof_size $proof_avgsize $total_time $proof_time $proof_avgtime" >> squolem2-statistics.txt
  done
fi

###############################################################################
# Creating a .tar.gz archive that contains the benchmarks, proofs, statistics
###############################################################################

archive="squolem2-certificates-$timeout.tar.gz"

# safety precaution: exit if $archive already exists
if [ -e "$archive" ]; then
  echo "File $archive exists, skipping archive creation."
  exit 1
fi

echo "Creating $archive ..."

tar -c * | gzip -9 > "$archive"

echo "... done."
